from typing import Any, Dict
from .sensor import Sensor
from .const import Identifier


'''
Data-Source:
https://www.sma.de/produkte/solar-wechselrichter/sunny-tripower-x
Technical Information - Parameters and Measured Values STP 12-50 / STP 15-50 / STP 20-50 / STP 25-50 (Sunny Tripower X) with firmware package 03.02.07.R 



'''
ennexMapper: Dict[int, str] = {
    35: "Error",
    51: "Closed",
    276: "Instantaneous value",
    295: "MPP",
    303: "Off",
    307: "OK",
    308: "On",
    309: "Operation",
    311: "Open",
    325: "Phase L1 (phsA)",
    326: "Phases L1, L2 and L3 (phsABC)",
    327: "Phase L2 (phsB)",
    329: "Phase L3 (phsC)",
    336: "Contact the manufacturer",
    337: "Contact the installer",
    338: "Invalid",
    381: "Stop",
    402: "Phases L1 and L2 (phsAB)",
    403: "Phases L1 and L3 (phsAC)",
    404: "Phases L2 and L3 (phsBC)",
    433: "Contant voltage",
    455: "Warning",
    461: "SMA",
    887: "No recommended action",
    1041: "leading / overexcited",
    1042: "lagging / underexcited",
    1069: "Reactive power / voltage characteristic curve Q(V)",
    1070: "Reactive power Q, direct setpoint",
    1071: "Reactive power const. Q (kVAr)",
    1072: "Q specified by PV system control",
    1073: "Reactive power Q(P)",
    1074: "cos φ, direct setpoint",
    1075: "cos φ, setpoint via system control",
    1076: "cos φ(P) characteristic curve",
    1077: "Active power limitation P (W)",
    1078: "Active power limitation P (%) of PMAX",
    1079: "Active power limitation P via system control",
    1129: "Yes",
    1130: "No",
    1264: "Full dynamic grid support",
    1265: "Limited dynamic grid support",
    1387: "Reactive power Q, setpoint via analog input",
    1388: "cos φ, setpoint via analog input",
    1389: "Reactive power / voltage characteristic curve Q(U) with hysteresis and deadband",
    1390: "Active power limitation P via analog input",
    1391: "Active power limitation P via digital inputs",
    1392: "Errors",
    1393: "Wait for PV voltage",
    1394: "Wait for valid AC grid",
    1395: "DC area",
    1396: "AC grid",
    1438: "Automatic",
    1455: "Emergency stop",
    1466: "Waiting",
    1467: "Starting",
    1468: "MPP search",
    1469: "Shutdown",
    1470: "Event message",
    1471: "Warning/error e-mail OK",
    1472: "Warning/error e-mail not OK",
    1473: "System info Parameter.Nameplate.Modele-mail OK",
    1474: "System info e-mail not OK",
    1475: "Error e-mail OK",
    1476: "Error e-mail not OK",
    1477: "Warning e-mail OK",
    1478: "Warning e-mail not OK",
    1479: "Wait after grid interruption",
    1480: "Wait for electric utility company",
    1749: "Full Stop",
    # 2055: "Status digital inlet: DI1",
    # 2056: "Status digital inlet: DI1, DI2",
    # 2057: "Status digital inlet: DI1, DI2, DI3",
    # 2058: "Status digital inlet: DI1, DI2, DI3, DI4",
    # 2059: "Status digital inlet: DI1, DI2, DI4",
    # 2060: "Status digital inlet: DI1, DI3",
    # 2061: "Status digital inlet: DI1, DI3, DI4",
    # 2062: "Status digital inlet: DI1, DI4",
    # 2063: "Status digital inlet: DI2",
    # 2064: "Status digital inlet: DI2, DI3",
    # 2065: "Status digital inlet: DI2, DI3, DI4",
    # 2066: "Status digital inlet: DI2, DI4",
    # 2067: "Status digital inlet: DI3",
    # 2068: "Status digital inlet: DI3, DI4",
    # 2069: "Status digital inlet: DI4",
    2270: "cos φ or Q specification through optimum PV system control",
    2506: "Values maintained",
    2507: "Apply fallback values",
    4354: "Maximum active power export",
    4405: "Maximum active power WMax",
    4406: "Maximum reactive power VArMax",
    4433: "Zero at dead band boundary",
    4434: "Zero at origin (ZerAtZer)",
    4443: "Current power",
    4444: "Potential power",
    4450: "Q limitation",
    4562: "cos φ(V) charac. curve",
    4520: "Mean value of the phase voltages",
    4521: "Maximum phase voltage",
    4718: "Boost Charging", # EV-Charger
    4950: "Smart Chargig", # EV-Charger
    5169: "Station locked", # EV-Charger
    5249: "Potential power with characteristic curve break",



    # DevClass
    8001: "Solar Inveters",
    8008: "SMAEVCharger",

    9489: "STP-20-50",
    9491: "STP-15-50",
    9492: "STP-12-50",
    9484: "EVC22-3AC-10",


    200111: "Not connected", # EV-Charger
    200112: "Sleep Mode", # EV-Charger
    200113: "Active Mode", # EV-Charger

    16777213: "Information not available",
}

ennexosSensorProfiles = {
    "Sunny Tripower X ": [
        Sensor("Coolsys.Inverter.TmpVal.1", Identifier.temp_a, factor=1, unit="°C"),
        Sensor("Coolsys.Inverter.TmpVal.2", Identifier.temp_b, factor=1, unit="°C"),
        Sensor("Coolsys.Inverter.TmpVal.3", Identifier.temp_c, factor=1, unit="°C"),
        Sensor("DcMs.Amp.1", Identifier.pv_current_a, factor=1, unit="A"),
        Sensor("DcMs.Amp.2", Identifier.pv_current_b, factor=1, unit="A"),
        Sensor("DcMs.Amp.3", Identifier.pv_current_c, factor=1, unit="A"),
        Sensor("DcMs.TotDcEnCntWh.1", None, factor=1, unit=None), # Energy released by string [A]
        Sensor("DcMs.TotDcEnCntWh.2", None, factor=1, unit=None),
        Sensor("DcMs.TotDcEnCntWh.3", None, factor=1, unit=None),
        Sensor("DcMs.Vol.1", Identifier.pv_voltage_a, factor=1, unit="V"),
        Sensor("DcMs.Vol.2", Identifier.pv_voltage_b, factor=1, unit="V"),
        Sensor("DcMs.Vol.3", Identifier.pv_voltage_c, factor=1, unit="V"),
        Sensor("DcMs.Watt.1", Identifier.pv_power_a, factor=1, unit="W"),
        Sensor("DcMs.Watt.2", Identifier.pv_power_b, factor=1, unit="W"),
        Sensor("DcMs.Watt.3", Identifier.pv_power_c, factor=1, unit="W"),
        Sensor("GridGuard.Cntry", None, factor=1, unit=None), # Country standard set
        Sensor("GridMs.A.phsA", Identifier.current_l1, factor=1, unit="A"),
        Sensor("GridMs.A.phsB", Identifier.current_l2, factor=1, unit="A"),
        Sensor("GridMs.A.phsC", Identifier.current_l3, factor=1, unit="A"),
        Sensor("GridMs.GriTyp", None, factor=1, unit=None), # Measurement.GridMs.GriTyp
        Sensor("GridMs.Hz", Identifier.frequency, factor=1, unit="Hz"),
        Sensor("GridMs.PhV.phsA", Identifier.voltage_l1, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsA2B", None, factor=1, unit=None), # Grid voltage phase L1 against L2
        Sensor("GridMs.PhV.phsB", Identifier.voltage_l2, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsB2C", None, factor=1, unit=None), # Grid voltage phase L2 against L3
        Sensor("GridMs.PhV.phsC", Identifier.voltage_l3, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsC2A", None, factor=1, unit=None), # Grid voltage phase L3 against L1
        Sensor("GridMs.TotA", Identifier.current_total, factor=1, unit="A"),
        Sensor("GridMs.TotPFEEI", None, factor=1, unit=None),  # EEI displacement power factor
        Sensor("GridMs.TotPFExt", None, factor=1, unit=None), # Excitation type of cos φ
        Sensor("GridMs.TotPFPrc", None, factor=1, unit=None), # Displacement power factor
        Sensor("GridMs.TotVA", Identifier.grid_apparent_power, factor=1, unit="VA"),
        Sensor("GridMs.TotVAr", Identifier.grid_reactive_power, factor=1, unit="var"),
        Sensor("GridMs.TotW", Identifier.grid_power, factor=1, unit="W"),
        Sensor("GridMs.TotW.Pv", Identifier.pv_power, factor=1, unit="W"),
        Sensor("GridMs.VA.phsA", Identifier.grid_apparent_power_l1, factor=1, unit="VA"),
        Sensor("GridMs.VA.phsB", Identifier.grid_apparent_power_l2, factor=1, unit="VA"),
        Sensor("GridMs.VA.phsC", Identifier.grid_apparent_power_l3, factor=1, unit="VA"),
        Sensor("GridMs.VAr.phsA", Identifier.grid_reactive_power_l1, factor=1, unit="var"),
        Sensor("GridMs.VAr.phsB", Identifier.grid_reactive_power_l2, factor=1, unit="var"),
        Sensor("GridMs.VAr.phsC", Identifier.grid_reactive_power_l3, factor=1, unit="var"),
        Sensor("GridMs.W.phsA", Identifier.power_l1, factor=1, unit="W"),
        Sensor("GridMs.W.phsB", Identifier.power_l2, factor=1, unit="W"),
        Sensor("GridMs.W.phsC", Identifier.power_l3, factor=1, unit="W"),
        Sensor("InOut.GI1", None, factor=1, unit=None), # Digital group input
        Sensor("Inverter.VArModCfg.PFCtlVolCfg.Stt", None, factor=1, unit=None),
        Sensor("Isolation.FltA", Identifier.insulation_residual_current, factor=1000, unit="mA"),
        Sensor("Isolation.LeakRis", None, factor=1, unit="kOhm"), # TODO "pv_isolation_resistance"
        Sensor("Metering.TotFeedTms", None, factor=1, unit=None), 
        Sensor("Metering.TotOpTms", None, factor=1, unit=None),
        Sensor("Metering.TotWhOut", Identifier.total_yield, factor=1000, unit="kWh"),
        Sensor("Metering.TotWhOut.Pv", Identifier.pv_gen_meter, factor=1000, unit="kWh"),
        Sensor("Operation.BckStt", None, factor=1, unit=None),
        Sensor("Operation.DrtStt", None, factor=1, unit=None),
        Sensor("Operation.Evt.Dsc", None, factor=1, unit=None),
        Sensor("Operation.Evt.EvtNo", None, factor=1, unit=None),
        Sensor("Operation.Evt.Msg", None, factor=1, unit=None),
        Sensor("Operation.EvtCntIstl", None, factor=1, unit=None),
        Sensor("Operation.EvtCntUsr", None, factor=1, unit=None),
        Sensor("Operation.GriSwCnt", None, factor=1, unit=None),
        Sensor("Operation.GriSwStt", None, factor=1, unit=None),
        Sensor("Operation.Health", Identifier.status, factor=1, unit=None, mapper=ennexMapper),
        Sensor("Operation.HealthStt.Alm", None, factor=1, unit=None),
        Sensor("Operation.HealthStt.Ok", None, factor=1, unit=None),
        Sensor("Operation.HealthStt.Wrn", None, factor=1, unit=None),
        Sensor("Operation.OpStt", None, factor=1, unit=None),
        Sensor("Operation.PvGriConn", None, factor=1, unit=None),
        Sensor("Operation.RstrLokStt", None, factor=1, unit=None),
        Sensor("Operation.RunStt", None, factor=1, unit=None),
        Sensor("Operation.StandbyStt", None, factor=1, unit=None),
        Sensor("Operation.VArCtl.VArModAct", None, factor=1, unit=None),
        Sensor("Operation.VArCtl.VArModStt", None, factor=1, unit=None),
        Sensor("Operation.WMaxLimSrc", None, factor=1, unit=None),
        Sensor("Operation.WMinLimSrc", None, factor=1, unit=None),
        Sensor("PvGen.PvW", None, factor=1, unit=None), #  PV generation power
        Sensor("PvGen.PvWh", None, factor=1, unit=None), # Meter count and PV gen. meter
        Sensor("Spdwr.ComSocA.Stt", None, factor=1, unit=None), # Speedwire connection status of SMACOM A
        Sensor("SunSpecSig.SunSpecTx.1", None, factor=1, unit=None), #SunSpec life sign [1]
        Sensor("Upd.Stt", None, factor=1, unit=None),
        Sensor("WebConn.Stt", None, factor=1, unit=None), #  Status of the Webconnect functionality
        Sensor("Wl.AcqStt", None, factor=1, unit=None), # Status of WiFi scan
        Sensor("Wl.AntMod", None, factor=1, unit=None), #  WiFi antenna type
        Sensor("Wl.ConnStt", None, factor=1, unit=None), # WiFi connection status
        Sensor("Wl.SigPwr", None, factor=1, unit=None), # Signal strength of the selected network
        Sensor("Wl.SoftAcsConnStt", None, factor=1, unit=None), #Soft Access Point status
        Sensor("Setpoint.PlantControl.InOut.GO1", None, factor=1, unit=None)
    ],
    "SMA EV Charger ": [
        Sensor("ChaSess.WhIn", None, factor=1, unit=None), # charging_session_energy
        Sensor("Chrg.ModSw", None, factor=1, unit=None, mapper=ennexMapper), # position_of_rotary_switch 4950 or 4718
        Sensor("GridMs.A.phsA", Identifier.current_l1, factor=1, unit="A"), #Netzstrom Phase L1
        Sensor("GridMs.A.phsB", Identifier.current_l2, factor=1, unit="A"),
        Sensor("GridMs.A.phsC", Identifier.current_l3, factor=1, unit="A"),
        Sensor("GridMs.Hz", Identifier.frequency, factor=1, unit="Hz"),
        Sensor("GridMs.PhV.phsA", Identifier.voltage_l1, factor=1, unit="V"),#Netzspannung Phase L1
        Sensor("GridMs.PhV.phsB", Identifier.voltage_l2, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsC", Identifier.voltage_l3, factor=1, unit="V"),
        Sensor("GridMs.TotPF", None, factor=1, unit=None),
        Sensor("GridMs.TotVA", Identifier.grid_apparent_power, factor=1, unit="VA"),
        Sensor("GridMs.TotVAr", Identifier.grid_reactive_power, factor=1, unit="var"),
        Sensor("InOut.GI1", None, factor=1, unit=None),
        Sensor("Metering.GridMs.TotWIn", Identifier.metering_power_absorbed, factor=1, unit="W"), #
        Sensor("Metering.GridMs.TotWIn.ChaSta", None, factor=1, unit=None), # same as Metering.GridMs.TotWIn
        Sensor("Metering.GridMs.TotWhIn", Identifier.metering_total_absorbed, factor=1, unit="Wh"), # charging_station_meter_reading
        Sensor("Metering.GridMs.TotWhIn.ChaSta", None, factor=1, unit=None), # same as Metering.GridMs.TotWhIn 
        Sensor("Operation.EVeh.ChaStt", Identifier.operating_status, factor=1, unit=None, mapper=ennexMapper), # charging_session_status
        Sensor("Operation.EVeh.Health", Identifier.status, factor=1, unit=None, mapper=ennexMapper), 
        Sensor("Operation.Evt.Msg", None, factor=1, unit=None),
        Sensor("Operation.Health", None, factor=1, unit=None), # same as EVeh.Health?
        Sensor("Operation.WMaxLimNom", None, factor=1, unit=None),
        Sensor("Operation.WMaxLimSrc", None, factor=1, unit=None),
        Sensor("Wl.AcqStt", None, factor=1, unit=None),
        Sensor("Wl.ConnStt", None, factor=1, unit=None),
        Sensor("Wl.SigPwr", None, factor=1, unit=None),
        Sensor("Wl.SoftAcsConnStt", None, factor=1, unit=None),
    ]
}



"""Test pysma init."""
import asyncio
import logging
from unittest.mock import patch
import json

import aiohttp
import pytest
import base64
import time
from pysma.device_em import SMAspeedwireEM
from pysma.exceptions import (
    SmaAuthenticationException,
    SmaConnectionException,
    SmaReadException,
)

from . import MOCK_DEVICE, MOCK_L10N, mock_aioresponse

_LOGGER = logging.getLogger(__name__)

class Test_SMAEM_class:
    """Test the SMA class."""

    def fake_recv(self):
        time.sleep(0.6)
        with open("tests/testdata/SunnyHomeManager2.json", "r") as file:
            data = json.load(file)
            return base64.b64decode(data["packet"])

    @patch.object(SMAspeedwireEM, '_recv', fake_recv)
    async def test_json_timeout_error(self, mock_aioresponse):
         sma = SMAspeedwireEM()
         device_info = await sma.device_info()
         print(device_info)
         for key in ["manufacturer", "name", "serial", "sw_version", "type" ]:
            assert key in device_info 
            assert len(device_info[key]) > 0 if isinstance(device_info[key], str) else True
            assert device_info[key] > 0 if isinstance(device_info[key], int) else True
         sensors = await sma.get_sensors()
         assert len(sensors) >= 17
         await sma.read(sensors)

    async def test_debug(self):
         sma = SMAspeedwireEM()
         with open("tests/testdata/SunnyHomeManager2.json", "r") as file:
            data = json.load(file)
            packet = base64.b64decode(data["packet"])
            sma._last_packet = packet
            debug = await sma.get_debug()
            assert debug["packet"] == data["packet"]  




from abc import ABC, abstractmethod
from typing import Dict
from .sensor import Sensors

class Device(ABC):

    @abstractmethod
    async def get_sensors(self) -> Sensors:
        pass
    
    @abstractmethod
    async def new_session(self) -> bool:
        pass

    @abstractmethod
    async def device_info(self) -> dict:
        pass

    @abstractmethod
    async def read(self, sensors: Sensors) -> bool:
        pass

    @abstractmethod
    async def close_session(self) -> None:
        pass

    @abstractmethod
    async def get_debug(self) -> Dict:
        pass




class Identifier():
    pv_power_a : str = "pv_power_a"
    pv_power_b : str = "pv_power_b"
    pv_power_c : str = "pv_power_c"
    pv_voltage_a : str = "pv_voltage_a"
    pv_voltage_b : str = "pv_voltage_b"
    pv_voltage_c : str = "pv_voltage_c"
    pv_current_a : str = "pv_current_a"
    pv_current_b : str = "pv_current_b"
    pv_current_c : str = "pv_current_c"
    insulation_residual_current : str = "insulation_residual_current"
    pv_power : str = "pv_power"
    grid_power : str = "grid_power"
    frequency : str = "frequency"
    power_l1 : str = "power_l1"
    power_l2 : str = "power_l2"
    power_l3 : str = "power_l3"
    temp_a : str = "temp_a"
    temp_b : str = "temp_b"
    temp_c : str = "temp_c"
    grid_reactive_power : str = "grid_reactive_power"
    grid_reactive_power_l1 : str = "grid_reactive_power_l1"
    grid_reactive_power_l2 : str = "grid_reactive_power_l2"
    grid_reactive_power_l3 : str = "grid_reactive_power_l3"
    grid_apparent_power : str = "grid_apparent_power"
    grid_apparent_power_l1 : str = "grid_apparent_power_l1"
    grid_apparent_power_l2 : str = "grid_apparent_power_l2"
    grid_apparent_power_l3 : str = "grid_apparent_power_l3"
    grid_power_factor : str = "grid_power_factor"
    current_l1 : str = "current_l1"
    current_l2 : str = "current_l2"
    current_l3 : str = "current_l3"
    current_total : str = "current_total"
    voltage_l1 : str = "voltage_l1"
    voltage_l2 : str = "voltage_l2"
    voltage_l3 : str = "voltage_l3"
    total_yield : str = "total_yield"
    daily_yield : str = "daily_yield"
    metering_power_supplied : str = "metering_power_supplied"
    metering_power_absorbed : str = "metering_power_absorbed"
    metering_frequency : str = "metering_frequency"
    metering_total_yield : str = "metering_total_yield"
    metering_total_absorbed : str = "metering_total_absorbed"
    metering_current_l1 : str = "metering_current_l1"
    metering_current_l2 : str = "metering_current_l2"
    metering_current_l3 : str = "metering_current_l3"
    metering_voltage_l1 : str = "metering_voltage_l1"
    metering_voltage_l2 : str = "metering_voltage_l2"
    metering_voltage_l3 : str = "metering_voltage_l3"
    metering_active_power_feed_l1 : str = "metering_active_power_feed_l1"
    metering_active_power_feed_l2 : str = "metering_active_power_feed_l2"
    metering_active_power_feed_l3 : str = "metering_active_power_feed_l3"
    metering_active_power_draw_l1 : str = "metering_active_power_draw_l1"
    metering_active_power_draw_l2 : str = "metering_active_power_draw_l2"
    metering_active_power_draw_l3 : str = "metering_active_power_draw_l3"
    metering_current_consumption : str = "metering_current_consumption"
    pv_gen_meter : str = "pv_gen_meter"
    sps_voltage : str = "sps_voltage"
    sps_current : str = "sps_current"
    sps_power : str = "sps_power"
    optimizer_serial : str = "optimizer_serial"
    optimizer_power : str = "optimizer_power"
    optimizer_current : str = "optimizer_current"
    optimizer_voltage : str = "optimizer_voltage"
    optimizer_temp : str = "optimizer_temp"
    battery_soc_total : str = "battery_soc_total"
    battery_soc_a : str = "battery_soc_a"
    battery_soc_b : str = "battery_soc_b"
    battery_soc_c : str = "battery_soc_c"
    battery_voltage_a : str = "battery_voltage_a"
    battery_voltage_b : str = "battery_voltage_b"
    battery_voltage_c : str = "battery_voltage_c"
    battery_current_a : str = "battery_current_a"
    battery_current_b : str = "battery_current_b"
    battery_current_c : str = "battery_current_c"
    battery_temp_a : str = "battery_temp_a"
    battery_temp_b : str = "battery_temp_b"
    battery_temp_c : str = "battery_temp_c"
    battery_capacity_total : str = "battery_capacity_total"
    battery_capacity_a : str = "battery_capacity_a"
    battery_capacity_b : str = "battery_capacity_b"
    battery_capacity_c : str = "battery_capacity_c"
    battery_charging_voltage_a : str = "battery_charging_voltage_a"
    battery_charging_voltage_b : str = "battery_charging_voltage_b"
    battery_charging_voltage_c : str = "battery_charging_voltage_c"
    battery_power_charge_total : str = "battery_power_charge_total"
    battery_power_charge_a : str = "battery_power_charge_a"
    battery_power_charge_b : str = "battery_power_charge_b"
    battery_power_charge_c : str = "battery_power_charge_c"
    battery_charge_total : str = "battery_charge_total"
    battery_charge_a : str = "battery_charge_a"
    battery_charge_b : str = "battery_charge_b"
    battery_charge_c : str = "battery_charge_c"
    battery_power_discharge_total : str = "battery_power_discharge_total"
    battery_power_discharge_a : str = "battery_power_discharge_a"
    battery_power_discharge_b : str = "battery_power_discharge_b"
    battery_power_discharge_c : str = "battery_power_discharge_c"
    battery_discharge_total : str = "battery_discharge_total"
    battery_discharge_a : str = "battery_discharge_a"
    battery_discharge_b : str = "battery_discharge_b"
    battery_discharge_c : str = "battery_discharge_c"
    serial_number : str = "serial_number"
    device_name : str = "device_name"
    device_type : str = "device_type"
    device_manufacturer : str = "device_manufacturer"
    device_sw_version : str = "device_sw_version"
    inverter_power_limit : str = "inverter_power_limit"
    energy_meter : str = "energy_meter"
    operating_status_genereal: str = "operating_status_general"
    operating_status: str = "operating_status"
    inverter_condition: str = "inverter_condition"
    inverter_system_init: str = "inverter_system_init"
    grid_connection_status: str = "grid_connection_status"
    grid_relay_status: str = "grid_relay_status"
    pv_isolation_resistance: str = "pv_isolation_resistance"
    grid_power_factor_excitation: str = "grid_power_factor_excitation"
    metering_total_consumption: str = "metering_total_consumption"
    battery_status_operating_mode: str = "battery_status_operating_mode"
    status: str = "status"



import asyncio
import copy
import json
import logging
from typing import Any, Dict, Optional


from aiohttp import ClientSession, ClientTimeout, client_exceptions, hdrs
from .sensor import Sensor
from .const_webconnect import (
    DEFAULT_TIMEOUT,
)
from .exceptions import (
    SmaAuthenticationException,
    SmaConnectionException,
    SmaReadException,
)
from .sensor import Sensors
from .device import Device
from .const import Identifier
from .definitions_ennexos import ennexosSensorProfiles

_LOGGER = logging.getLogger(__name__)
class SMAennexos(Device):
    """Class to connect to the ennexos based SMA inverters. (e.g. Tripower X Devices)"""

    # pylint: disable=too-many-instance-attributes
    _aio_session: ClientSession
    _new_session_data: Optional[dict]
    _url: str
    _token: str
    _authorization_header: str
    _last_parameters: Any
    _last_measurements: Any
    _last_device: Any
    _device_info: Dict = None
    _jsessionid: str = None    
    def __init__(
        self,
        session: ClientSession,
        url: str,
        password: Optional[str],
        group: str,
    ):
        """Init SMA connection.

        Args:
            session (ClientSession): aiohttp client session
            url (str): Url or IP address of device
            password (str, optional): Password to use during login.
            group (str, optional): Username to use during login. 

        """
        self._url = url.rstrip("/")
        if not url.startswith("http"):
            self._url = "https://" + self._url
        self._new_session_data = {"user": group, "pass": password}
        self._aio_session = session



    async def _jsonrequest(
        self, url: str, parameters: Dict[str, Any], method: str = hdrs.METH_POST
    ) -> dict:
        """Request json data for requests.

        Args:
            url (str): URL to do request to
            parameters (Dict[str, Any]): parameters
            method (str): Post or Get-Request

        Raises:
            SmaConnectionException: Connection to device failed
            SmaAuthenticationException: Authentication failed

        Returns:
            dict: json returned by device
        """

        #_LOGGER.debug("Sending Request to %s: %s", url, parameters)

        try:
                async with self._aio_session.request(
                    method,
                    url,
                    timeout=ClientTimeout(total=DEFAULT_TIMEOUT),
                    **parameters
                ) as res:
                    if (res.status == 401):
                        raise SmaAuthenticationException(
                            "Token failed!"
                        )
                    res = await res.json()
                   # _LOGGER.debug("Received reply %s", res)
                    return res
        except SmaAuthenticationException as e:
                raise e
        except (client_exceptions.ContentTypeError, json.decoder.JSONDecodeError):
            _LOGGER.warning("Request to %s did not return a valid json.", url)
        except client_exceptions.ServerDisconnectedError as exc:
            raise SmaConnectionException(
                f"Server at {self._url} disconnected."
            ) from exc
        except (
            client_exceptions.ClientError,
            asyncio.exceptions.TimeoutError,
        ) as exc:
            raise SmaConnectionException(
                f"Could not connect to SMA at {self._url}: {exc}"
            ) from exc
        return {}


    async def new_session(self) -> bool:
        """Establish a new session.

        Returns:
            bool: authentication successful
        """
        loginurl = self._url + '/api/v1/token'
        postdata = {'data':{'grant_type': 'password',
                'username': self._new_session_data["user"],
                'password': self._new_session_data["pass"],
                }}
        ret = await self._jsonrequest(loginurl,postdata)
        if "access_token" not in ret:
            raise SmaAuthenticationException("Login failed!")
        self._token = ret["access_token"]
#        self._refreshtoken = ret["refresh_token"]
        self._authorization_header = { "Authorization" : "Bearer " + self._token,
                                       "Content-Type": "application/json"
                                       }
        _LOGGER.debug("Login successfull")
        return True

    async def _get_parameter(self) -> Dict:
        """Get all parameters from the device

        Returns:
            Dict: Return a dict with all parameters

        """
        url = self._url + '/api/v1/parameters/search'
        postdata = {
             'data': '{"queryItems":[{"componentId":"IGULD:SELF"}]}',
             'headers': self._authorization_header
        }
        ret = await self._jsonrequest(url, postdata)
        data = {}
        if (len(ret) != 1):
            _LOGGER.warning("Uncommon length of array in parameters request: %d", len(ret))

        for d in ret[0]["values"]:
            dname = d["channelId"].replace("Parameter.","").replace("[]", "")
            if "value" in d:
                v = d["value"]
                data[dname] = {
                        "name": dname,
                        "value": v,
                        "origname": d["channelId"]
                    }
            elif "values" in d:
                # Split Value-Arrays
                for idx in range(0, len(d["values"])):
                    v = d["values"][idx]
                    idxname = dname + "." + str(idx + 1)
                    data[idxname] = {
                            "name": idxname,
                            "value": v,
                            "origname": d["channelId"]
                    }
            else:
                # Value current not available // night?
                # TODO
                pass
        return data

    async def _get_livedata(self) -> Dict:
        """Get the sensors reading from the device

        Returns:
            Dict: Return a dict with all measurements

        """
        liveurl = self._url + '/api/v1/measurements/live'
        postdata = { 
             'data': '[{"componentId":"IGULD:SELF"}]',
             'headers': self._authorization_header
        }
        ret = await self._jsonrequest(liveurl,postdata)
        self._last_measurements = ret
        data = {}
        for d in ret:
            dname = d["channelId"].replace("Measurement.","").replace("[]", "")
            if "value" in d["values"][0]:
                v = d["values"][0]["value"]
                if self._isfloat(v):
                     v = round(v,2)
                data[dname] = {
                        "name": dname,
                        "value": v,
                        "origname": d["channelId"]
                    }
            elif "values" in d["values"][0]:
                # Split Value-Arrays
                for idx in range(0, len(d["values"][0]["values"])):
                    v = d["values"][0]["values"][idx]
                    if self._isfloat(v):
                        v = round(v,2)
                    idxname = dname + "." + str(idx + 1)
                    data[idxname] = {
                            "name": idxname,
                            "value": v,
                            "origname": d["channelId"]
                    }
            else:
                # Value current not available // night?
                # TODO
                pass
        return data
    


    async def get_sensors(self) -> Sensors:
        """Get the sensors that are present on the device.

        Returns:
            Sensors: Sensors object containing Sensor objects
        """
        if not self._device_info:
            raise SmaReadException("device_info() not called!")

        ret = await self._get_livedata()
        _LOGGER.debug("Found Sensors: %s", ret)

        device_sensors = Sensors()
        sensors = []

        # Search for matiching profile
        for devname in ennexosSensorProfiles.keys():
            if (self._device_info["name"].startswith(devname)):
                sensors = ennexosSensorProfiles[devname]
        if (len(sensors) == 0):
            _LOGGER.warning(f'Unknown Device: {self._device_info["name"]} {self._device_info["type"]}')
        
        # Add Sensors from profile
        for s in sensors:
             if s.name:
                 device_sensors.add(copy.copy(s))
        return device_sensors


    async def close_session(self) -> None:
        pass

    def _isfloat(self,num: Any):
        """Test if num is a float
            
            Tests for type float or a string with a dot is is float
        
        Args:
            num: number to check

        Returns:
            bool: true, if num is from type float or a string with a dot
        """
        if isinstance(num, float):
            return True
        if isinstance(num, int):
            return False
        if not isinstance(num, str):
            raise TypeError("Value is not a string, float or int!")
        if ("." not in num):
            return False
        try:
            float(num)
            return True
        except ValueError:
            return False


    async def read(self, sensors: Sensors) -> bool:
        """Read a set of keys.

        Args:
            sensors (Sensors): Sensors object containing Sensor objects to read

        Returns:
            bool: reading was successful
        """
        notfound = []
        data = None
        try:
            data = await self._get_livedata()
        except SmaAuthenticationException:
            # Relogin
            _LOGGER.debug("Re-login .. Starting new Session")
            await self.new_session()
            data = await self._get_livedata()

        for sen in sensors:
            if sen.enabled:
                if sen.key in data:
                    value = data[sen.key]["value"]
                    if (sen.mapper):
                        sen.mapped_value = sen.mapper.get(value, str(value))
                    if (sen.factor and sen.factor != 1):
                        value = round(value/sen.factor,4)
                    sen.value = value
                    continue
                notfound.append(f"{sen.name} [{sen.key}]")

        if notfound:
            _LOGGER.info(
                "No values for sensors: %s",
                ",".join(notfound),
            )

        return True



    async def device_info(self) -> dict:
        """Read device info and return the results.

        Returns:
            dict: dict containing serial, name, type, manufacturer and sw_version
        """
        url = self._url + '/api/v1/plants/Plant:1/devices/IGULD:SELF'
        requestdata = { 
             'headers': self._authorization_header
        }
        dev = await self._jsonrequest(url,requestdata, hdrs.METH_GET)
        self._last_device = dev
        self._last_parameters = await self._get_parameter()
        _LOGGER.debug("Found Device: %s", dev)
        self._device_info = {
            "serial": dev["serial"],
            "name": dev["product"],
            "type": dev["name"],
            "manufacturer": dev["vendor"],
            "sw_version": dev["firmwareVersion"],
        }
        return self._device_info

    async def get_debug(self) -> Dict:
        """ Collect all Debug Information """
        return {
            "device": self._last_device,
            "measurements": self._last_measurements,
            "parameters": self._last_parameters
        }




"""SMA WebConnect library for Python.

See: http://www.sma.de/en/products/monitoring-control/webconnect.html

Source: http://www.github.com/kellerza/pysma
"""
import logging
from typing import Optional

from aiohttp import ClientSession
from .device_webconnect import SMAwebconnect
from .device_ennexos import SMAennexos
from .device_em import SMAspeedwireEM
from .device import Device
_LOGGER = logging.getLogger(__name__)


# Backward compatibility
def SMA(session, url, password, group):
    return SMAwebconnect(session, url, password=password, group=group)


def getDevice(session: ClientSession,
        url: str,
        password: Optional[str] = None,
        groupuser: str = "user",
        accessmethod: str = "webconnect"
    ) -> Device:
        _LOGGER.debug(f"Device Called! Url: {url} User/Group: {groupuser} Accessmethod: {accessmethod}")
        if (accessmethod == "webconnect"):
            return SMAwebconnect(session, url, password=password, group=groupuser)
        elif (accessmethod == "ennexos"):
            return SMAennexos(session, url, password=password, group=groupuser)
        elif (accessmethod == "speedwire"):
              return SMAspeedwireEM()
        else:
             return None



import base64
import socket
import struct
import copy
import logging
from typing import Any, Dict
import datetime

# https://www.unifox.at/software/sma-em-daemon/
# https://cdn.sma.de/fileadmin/content/www.developer.sma.de/docs/EMETER-Protokoll-TI-en-10.pdf?v=1699276024


from .sensor import Sensor
from .const import Identifier

from .exceptions import (
    SmaConnectionException,
    SmaReadException,
)
from .sensor import Sensors
from .device import Device

obis2sensor= [
    Sensor("1:4:0", Identifier.metering_power_absorbed, factor=10, unit="W"), # p consume
    Sensor("1:8:0", Identifier.metering_total_absorbed, factor=3600000, unit="kWh"),
    Sensor("2:4:0", Identifier.metering_power_supplied, factor=10, unit="W"), # p supply
    Sensor("2:8:0", Identifier.metering_total_yield, factor=3600000, unit="kWh", ),
    Sensor("3:4:0", None), # q consume
    Sensor("3:8:0", None),
    Sensor("4:4:0", None), # q supply
    Sensor("4:8:0", None),
    Sensor("9:4:0", None), # s consume
    Sensor("9:8:0", None),
    Sensor("10:4:0", None), # s supply
    Sensor("10:8:0", None),
    Sensor("13:4:0", None), # cospi
    Sensor("14:4:0", Identifier.metering_frequency, factor=1000, unit="Hz"), # freq

    # Phase 1

    Sensor("21:4:0", Identifier.metering_active_power_draw_l1, factor=10, unit="W"),
    Sensor("21:8:0", None),
    Sensor("22:4:0", Identifier.metering_active_power_feed_l1, factor=10, unit="W"),
    Sensor("22:8:0", None),
    Sensor("23:4:0", None),
    Sensor("23:8:0", None),
    Sensor("24:4:0", None),
    Sensor("24:8:0", None),
    Sensor("29:4:0", None),
    Sensor("29:8:0", None),
    Sensor("30:4:0", None),
    Sensor("30:8:0", None),
    Sensor("31:4:0", Identifier.metering_current_l1, factor=1000, unit="A"),
    Sensor("32:4:0", Identifier.metering_voltage_l1, factor=1000, unit="V"),
    Sensor("33:4:0", None), #cosphi1

    # Phase 2

    Sensor("41:4:0", Identifier.metering_active_power_draw_l2, factor=10, unit="W"),
    Sensor("41:8:0", None),
    Sensor("42:4:0", Identifier.metering_active_power_feed_l2, factor=10, unit="W"),
    Sensor("42:8:0", None),
    Sensor("43:4:0", None),
    Sensor("43:8:0", None),
    Sensor("44:4:0", None),
    Sensor("44:8:0", None),
    Sensor("49:4:0", None),
    Sensor("49:8:0", None),
    Sensor("50:4:0", None),
    Sensor("50:8:0", None),
    Sensor("51:4:0", Identifier.metering_current_l2, factor=1000, unit="A"),
    Sensor("52:4:0", Identifier.metering_voltage_l2, factor=1000, unit="V"),
    Sensor("53:4:0", None),

    # Phase 3
    Sensor("61:4:0", Identifier.metering_active_power_draw_l3, factor=10, unit="W"),
    Sensor("61:8:0", None),
    Sensor("62:4:0", Identifier.metering_active_power_feed_l3, factor=10, unit="W"),
    Sensor("62:8:0", None),
    Sensor("63:4:0", None),
    Sensor("63:8:0", None),
    Sensor("64:4:0", None),
    Sensor("64:8:0", None),
    Sensor("69:4:0", None),
    Sensor("69:8:0", None),
    Sensor("70:4:0", None),
    Sensor("70:8:0", None),
    Sensor("71:4:0", Identifier.metering_current_l3, factor=1000, unit="A"),
    Sensor("72:4:0", Identifier.metering_voltage_l3, factor=1000, unit="V"),
    Sensor("73:4:0", None)
]


_LOGGER = logging.getLogger(__name__)
class SMAspeedwireEM(Device):
    """Class to connect to the ennexos based SMA inverters. (e.g. Tripower X Devices)"""
    _sock: socket
    _susyid: Dict[int, Any] = {270: "Energy Meter", 349: "Energy Meter 2", 372: "Sunny Home Manager 2"}
    _last_packet: bytes = None

    def __init__(self):
        """Init SMA connection.

        Args:
            session (ClientSession): aiohttp client session
            url (str): Url or IP address of device
            password (str, optional): Password to use during login.
            group (str, optional): Username to use during login. 

        """
        pass

    async def new_session(self) -> bool:
        """Establish a new session.

        Returns:
            bool: authentication successful
        """
        MCAST_GRP = '239.12.255.254'
        MCAST_PORT = 9522
        IPBIND = '0.0.0.0'

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.settimeout(5)
        self._sock.bind(("", MCAST_PORT))
        try:
            mreq = struct.pack("4s4s", socket.inet_aton(MCAST_GRP), socket.inet_aton(IPBIND))
            self._sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        except BaseException as exc:
            raise SmaConnectionException(
                "Could not start multicast"
            ) from exc


        return True


    async def get_sensors(self) -> Sensors:
        """Get the sensors that are present on the device.

        Returns:
            Sensors: Sensors object containing Sensor objects
        """
        device_sensors = Sensors()

        for s in obis2sensor:
            if s.name is not None:
                device_sensors.add(copy.copy(s)) 

        return device_sensors


    async def close_session(self) -> None:
        """Close the session login."""
        self._sock.close()

    def _recv(self):
        return self._sock.recv(608)

    def _getData(self):
        """

        Hack:
            If the function is called less frequently than the device supplies data,
            the UDP packets will be stored in the buffer.
            If the read instruction (sock.recv) takes only milliseconds, I assume that
            the data came from the buffer and discard it. Otherwise outdated values would
            be used.
            One of my most ugly hacks.

        """
        data = None
        tries = 50
        try:
            while tries > 0:
                tries -= 1
                a = datetime.datetime.now()
                self._last_packet = self._recv()
                data=self._decode(self._last_packet)
                b = datetime.datetime.now()
                if data and (b-a).total_seconds() < 0.1:
                    continue
                if data:
                    break
        except TimeoutError as e:
            raise SmaConnectionException("No speedwire packet received!") from e
        if not data:
            raise SmaReadException("No usable data received!")
        return data

    async def read(self, sensors: Sensors) -> bool:
        """Read a set of keys.

        Args:
            sensors (Sensors): Sensors object containing Sensor objects to read

        Returns:
            bool: reading was successful
        """
        notfound = []
        data = self._getData()
            
        for sensor in sensors:
          if (sensor.key in data):
              value = data[sensor.key]
              if (sensor.factor):
                  value /= sensor.factor
              sensor.value = value
          else:
              notfound.append(sensor.key)

        if notfound:
            _LOGGER.info(
                "No values for sensors: %s",
                ",".join(notfound),
            )

        return True


    async def device_info(self) -> dict:
        """Read device info and return the results.

        Returns:
            dict: dict containing serial, name, type, manufacturer and sw_version
        """
        data = self._getData()

        device_info = {
            "serial": data["serial"],
            "name": data["device"],
            "type": data["susyid"],
            "manufacturer": "SMA",
            "sw_version": data["sw_version"],
        }
        return device_info

    def _decode(self, p : bytes):
        """Decode a Speedwire-Packet

        Args:
            p: Network-Packet

        Returns:
            dict: Dict with all the decoded information
        """
        if p[0:4] != b"SMA\0":
            return None
        protocolID = int.from_bytes(p[16:18], byteorder="big")
        
        if (protocolID not in [0x6069, 0x6081]):
            _LOGGER.debug("Unknown protocoll " + str(protocolID))
            return None
        
        data = {}
        data["protocolID"] = protocolID
        data["susyid"] = int.from_bytes(p[18:20], byteorder="big")
        data["device"] = self._susyid.get(data["susyid"], "unknown")
        data["serial"] = int.from_bytes(p[20:24], byteorder="big")

        length = int.from_bytes(p[12:14], byteorder="big") + 16
        pos = 28
        while pos < length:
            value = None
            mchannel = int.from_bytes(p[pos: pos + 1], byteorder="big")
            mvalueindex = int.from_bytes(p[pos + 1: pos + 2], byteorder="big")
            mtyp = int.from_bytes(p[pos + 2: pos + 3], byteorder="big")
            mtariff = int.from_bytes(p[pos + 3: pos + 4], byteorder="big")
            obis = f'{mvalueindex}:{mtyp}:{mtariff}'
            if mtyp in [4,8]:
                # 4 actucal / current => 8 Bytes
                # 8 counter / sum => 12 Bytes
                value = int.from_bytes(p[pos + 4 : pos + 4 + mtyp], byteorder="big")
                pos += 4 + mtyp
            elif mchannel == 144 and mtyp == 0:
                value = f"{p[pos + 4]}.{p[pos + 5]}.{p[pos + 6]}.{chr(p[pos + 7])}"
                obis= "sw_version"
                pos += 4 + 4
            else:
                _LOGGER.debug("Unknown packet in speedwire: " + str(mchannel) + " " + str(mvalueindex) + " " + str(mtyp) + " " + str(mtariff))
                pos += 4 + 4
            data[obis] = value
        return data

    async def get_debug(self) -> Dict:
        encoded = base64.b64encode(self._last_packet)
        return { "packet": encoded.decode('ascii')}



"""Constants for SMA WebConnect library for Python."""
URL_LOGIN = "/dyn/login.json"
URL_LOGOUT = "/dyn/logout.json"
URL_VALUES = "/dyn/getValues.json"
URL_ALL_VALUES = "/dyn/getAllOnlValues.json"
URL_ALL_PARAMS = "/dyn/getAllParamValues.json"
URL_LOGGER = "/dyn/getLogger.json"
URL_DASH_LOGGER = "/dyn/getDashLogger.json"
URL_DASH_VALUES = "/dyn/getDashValues.json"

USERS = {"user": "usr", "installer": "istl"}

DEFAULT_TIMEOUT = 15
DEFAULT_LANG = "en-US"

JMESPATH_VAL = "val"
JMESPATH_VAL_TAG = JMESPATH_VAL + "[0].tag"
JMESPATH_VAL_STR = "[?str==sum([`1`,`{}`])].val | [0]"
JMESPATH_VAL_IDX = "* | [0][{}].val"
JMESPATH_VAL_IDX_TAG = JMESPATH_VAL_IDX + "[0].tag"

JMESPATHS_TAG = (JMESPATH_VAL_IDX_TAG, JMESPATH_VAL_TAG)

LEGACY_MAP = {
    "pv_power": {"old_key": "6100_0046C200", "new_sensor": "pv_power_a"},
    "pv_voltage": {
        "old_key": "6380_40451F00",
        "new_sensor": "pv_power_a",
    },
    "pv_current": {
        "old_key": "6380_40452100",
        "new_sensor": "pv_current_a",
    },
}

FALLBACK_DEVICE_INFO = {
    "manufacturer": "SMA",
    "name": "SMA Device",
    "type": "",
    "serial": "9999999999",
}

GENERIC_SENSORS = "generic"
OPTIMIZERS_VIA_INVERTER = "optimizers"
ENERGY_METER_VIA_INVERTER = "energy-meter"
DEVICE_INFO = "device-info"

